# The main simulator for serverless computing platforms

from pacssim.SimProcess import ExpSimProcess
from pacssim.FunctionInstance import FunctionInstance
import numpy as np
import pandas as pd

from tqdm import tqdm

class ServerlessSimulator:
    """ServerlessSimulator is responsible for executing simulations of a sample serverless computing platform, mainly for the performance analysis and performance model evaluation purposes.

    Parameters
    ----------
    arrival_process : pacssim.SimProcess.SimProcess, optional
        The process used for generating inter-arrival samples, if absent, `arrival_rate` should be passed to signal exponential distribution, by default None
    warm_service_process : pacssim.SimProcess.SimProcess, optional
        The process which will be used to calculate service times, if absent, `warm_service_rate` should be passed to signal exponential distribution, by default None
    cold_service_process : pacssim.SimProcess.SimProcess, optional
        The process which will be used to calculate service times, if absent, `cold_service_rate` should be passed to signal exponential distribution, by default None
    expiration_threshold : float, optional
        The period of time after which the instance will be expired and the capacity release for use by others, by default 600
    max_time : float, optional
        The maximum amount of time for which the simulation should continue, by default 24*60*60 (24 hours)
    maximum_concurrency : int, optional
        The maximum number of concurrently executing function instances allowed on the system This will be used to determine when a rejection of request should happen due to lack of capacity, by default 1000

    Raises
    ------
    Exception
        Raises if neither arrival_process nor arrival_rate are present
    Exception
        Raises if neither warm_service_process nor warm_service_rate are present
    Exception
        Raises if neither cold_service_process nor cold_service_rate are present
    ValueError
        Raises if warm_service_rate is smaller than cold_service_rate
    """
    def __init__(self, arrival_process=None, warm_service_process=None, 
            cold_service_process=None, expiration_threshold=600, max_time=24*60*60,
            maximum_concurrency=1000, **kwargs):
        super().__init__()
        
        # setup arrival process
        self.arrival_process = arrival_process
        # if the user wants a exponentially distributed arrival process
        if 'arrival_rate' in kwargs:
            self.arrival_process = ExpSimProcess(rate=kwargs.get('arrival_rate'))
        # in the end, arrival process should be defined
        if self.arrival_process is None:
            raise Exception('Arrival process not defined!')

        # if both warm and cold service rate is provided (exponential distribution)
        # then, warm service rate should be larger than cold service rate
        if 'warm_service_rate' in kwargs and 'cold_service_rate' in kwargs:
            if kwargs.get('warm_service_rate') < kwargs.get('cold_service_rate'):
                raise ValueError("Warm service rate cannot be smaller than cold service rate!")

        # setup warm service process
        self.warm_service_process = warm_service_process
        if 'warm_service_rate' in kwargs:
            self.warm_service_process = ExpSimProcess(rate=kwargs.get('warm_service_rate'))
        if self.warm_service_process is None:
            raise Exception('Warm Service process not defined!')

        # setup cold service process
        self.cold_service_process = cold_service_process
        if 'cold_service_rate' in kwargs:
            self.cold_service_process = ExpSimProcess(rate=kwargs.get('cold_service_rate'))
        if self.cold_service_process is None:
            raise Exception('Cold Service process not defined!')

        self.expiration_threshold = expiration_threshold
        self.max_time = max_time
        self.maximum_concurrency = maximum_concurrency

        # reset trace values
        self.reset_trace()

    def reset_trace(self):
        """resets all the historical data to prepare the class for a new simulation
        """
        # an archive of previous servers
        self.prev_servers = []
        self.total_req_count = 0
        self.total_cold_count = 0
        self.total_warm_count = 0
        self.total_reject_count = 0
        # current state of instances
        self.servers = []
        self.server_count = 0
        self.running_count = 0
        self.idle_count = 0
        # history results
        self.hist_times = []
        self.hist_server_count = []
        self.hist_server_running_count = []
        self.hist_server_idle_count = []
        self.hist_req_cold_idxs = []
        self.hist_req_warm_idxs = []
        self.hist_req_rej_idxs = []

    def has_server(self):
        """Returns True if there are still instances (servers) in the simulated platform, False otherwise.

        Returns
        -------
        bool
            Whether or not the platform has instances (servers)
        """
        return len(self.servers) > 0

    def __str__(self):
        return f"idle/running/total: \t {self.idle_count}/{self.running_count}/{self.server_count}"

    def req(self):
        """Generate a request inter-arrival from `self.arrival_process`

        Returns
        -------
        float
            The generated inter-arrival sample
        """
        return self.arrival_process.generate_trace()

    def cold_start_arrival(self, t):
        """Goes through the process necessary for a cold start arrival which includes generation of a new function instance in the `COLD` state and adding it to the cluster.

        Parameters
        ----------
        t : float
            The time at which the arrival has happened. This is used to record the creation time for the server and schedule the expiration of the instance if necessary.
        """
        self.total_req_count += 1

        # reject request if maximum concurrency reached
        if self.running_count == self.maximum_concurrency:
            self.total_reject_count += 1
            self.hist_req_rej_idxs.append(len(self.hist_times) - 1)
            return

        self.total_cold_count += 1
        self.hist_req_cold_idxs.append(len(self.hist_times) - 1)

        self.server_count += 1
        self.running_count += 1
        new_server = FunctionInstance(t, self.cold_service_process, self.warm_service_process, self.expiration_threshold)
        self.servers.append(new_server)

    def schedule_warm_instance(self, t):
        """Goes through a process to determine which warm instance should process the incoming request.

        Parameters
        ----------
        t : float
            The time at which the scheduling is happening

        Returns
        -------
        pacssim.FunctionInstance.FunctionInstance
            The function instances that the scheduler has selected for the incoming request.
        """
        idle_instances = [s for s in self.servers if s.is_idle()]
        creation_times = [s.creation_time for s in idle_instances]
        
        # scheduling mechanism
        creation_times = np.array(creation_times)
        # find the newest instance
        idx = np.argmax(creation_times)
        return idle_instances[idx]

    def warm_start_arrival(self, t):
        """Goes through the process necessary for a warm start arrival which includes selecting a warm instance for processing and recording the request information.

        Parameters
        ----------
        t : float
            The time at which the arrival has happened. This is used to record the creation time for the server and schedule the expiration of the instance if necessary.
        """
        self.total_req_count += 1

        # reject request if maximum concurrency reached
        if self.running_count == self.maximum_concurrency:
            self.total_reject_count += 1
            self.hist_req_rej_idxs.append(len(self.hist_times) - 1)
            return

        self.hist_req_warm_idxs.append(len(self.hist_times) - 1)

        # schedule the request
        instance = self.schedule_warm_instance(t)
        instance.arrival_transition(t)

        # transition from idle to running
        self.total_warm_count += 1
        self.idle_count -= 1
        self.running_count += 1

    def get_trace_end(self):
        """Get the time at which the trace (one iteration of the simulation) has ended. This mainly due to the fact that we keep on simulating until the trace time goes beyond max_time, but the time is incremented until the next event.

        Returns
        -------
        float
            The time at which the trace has ended
        """
        return self.hist_times[-1]

    def calculate_time_lengths(self):
        """Calculate the time length for each step between two event transitions. Records the values in `self.time_lengths`.
        """
        self.time_lengths = np.diff(self.hist_times)

    def get_average_server_count(self):
        """Get the time-average server count.

        Returns
        -------
        float
            Average server count
        """
        avg_server_count = (self.hist_server_count * self.time_lengths).sum() / self.get_trace_end()
        return avg_server_count

    def get_average_server_running_count(self):
        """Get the time-averaged running server count.

        Returns
        -------
        float
            Average running server coutn
        """
        avg_running_count = (self.hist_server_running_count * self.time_lengths).sum() / self.get_trace_end()
        return avg_running_count

    def get_average_server_idle_count(self):
        """Get the time-averaged idle server count.

        Returns
        -------
        float
            Average idle server coutn
        """
        avg_idle_count = (self.hist_server_idle_count * self.time_lengths).sum() / self.get_trace_end()
        return avg_idle_count

    def get_index_after_time(self, t):
        """Get the first historical array index (for all arrays storing hisotrical events) that is after the time t.

        Parameters
        ----------
        t : float
            The time in the beginning we want to skip

        Returns
        -------
        int
            The calculated index in `self.hist_times`
        """
        return np.min(np.where(np.array(self.hist_times) > t))

    def get_skip_init(self, skip_init_time=None, skip_init_index=None):
        """Get the minimum index which satisfies both the time and index count we want to skip in the beginning of the simulation, which is used to reduce the transient effect for calculating the steady-state values.

        Parameters
        ----------
        skip_init_time : float, optional
            The amount of time skipped in the beginning, by default None
        skip_init_index : [type], optional
            The number of indices we want to skip in the historical events, by default None

        Returns
        -------
        int
            The number of indices after which both index and time requirements are satisfied
        """
        # how many initial values should be skipped
        skip_init = 0
        if skip_init_time is not None:
            skip_init = self.get_index_after_time(skip_init_time)
        if skip_init_index is not None:
            skip_init = max(skip_init, skip_init_index)
        return skip_init

    def get_request_custom_states(self, hist_states, skip_init_time=None, skip_init_index=None):
        """Get request statistics for an array of custom states.

        Parameters
        ----------
        hist_states : list[object]
            An array of custom states calculated by the user for which the statistics should be calculated, should be the same size as `hist_*` objects, these values will be used as the keys for the returned dataframe.
        skip_init_time : float, optional
            The amount of time skipped in the beginning, by default None
        skip_init_index : int, optional
            The number of indices that should be skipped in the beginning to calculate steady-state results, by default None

        Returns
        -------
        pandas.DataFrame
            A pandas dataframe including different statistics like `p_cold` (probability of cold start)
        """
        req_skip_init = self.get_skip_init(skip_init_time=skip_init_time, 
                                        skip_init_index=skip_init_index)

        state_req_colds = {}
        state_req_warm = {}
        state_req_rejs = {}
        for s in hist_states[req_skip_init:]:
            if s not in state_req_colds:
                state_req_colds[s] = 0
                state_req_warm[s] = 0
                state_req_rejs[s] = 0

        hist_req_cold = [i for i in self.hist_req_cold_idxs if i > req_skip_init]
        hist_req_warm = [i for i in self.hist_req_warm_idxs if i > req_skip_init]
        hist_req_rej = [i for i in self.hist_req_rej_idxs if i > req_skip_init]


        for idx in hist_req_cold:
            state_req_colds[hist_states[idx]] += 1
        for idx in hist_req_warm:
            state_req_warm[hist_states[idx]] += 1
        for idx in hist_req_rej:
            state_req_warm[hist_states[idx]] += 1

        states = list(state_req_colds.keys())
        state_req_colds = list(state_req_colds.values())
        state_req_warm = list(state_req_warm.values())
        state_req_rejs = list(state_req_rejs.values())

        reqdf = pd.DataFrame(data = {'state': states, 'cold': state_req_colds, 'warm': state_req_warm, 'rej': state_req_rejs})
        reqdf['total'] = reqdf['cold'] + reqdf['warm'] + reqdf['rej']
        reqdf['p_cold'] = reqdf['cold'] / reqdf['total']
        return reqdf

    def analyze_custom_states(self, hist_states, skip_init_time=None, skip_init_index=None):
        """Analyses a custom states list and calculates the amount of time spent in each state each time we enterred that state, and the times at which transitions have happened.

        Parameters
        ----------
        hist_states : list[object]
            The states calculated, should have the same dimensions as the `hist_*` arrays.
        skip_init_time : float, optional
            The amount of time skipped in the beginning, by default None
        skip_init_index : int, optional
            The number of indices skipped in the beginning, by default None

        Returns
        -------
        list[float], list[float]
            (residence_times, transition_times) where residence_times is an array of the amount of times we spent in each state, and transition_times are the moments of time at which each transition has occured
        """
        skip_init = self.get_skip_init(skip_init_time=skip_init_time, 
                                        skip_init_index=skip_init_index)

        values = hist_states[skip_init:]
        time_lengths = self.time_lengths[skip_init:]

        residence_times = {}
        transition_times = {}
        curr_time_sum = time_lengths[0]
        # encode states
        for idx in range(1,len(values)):
            if values[idx] == values[idx-1]:
                curr_time_sum += time_lengths[idx]
            else:
                if values[idx-1] in residence_times:
                    residence_times[values[idx-1]].append(curr_time_sum)
                else:
                    residence_times[values[idx-1]] = [curr_time_sum]

                transition_pair = (values[idx-1], values[idx])
                if transition_pair in transition_times:
                    transition_times[transition_pair].append(curr_time_sum)
                else:
                    transition_times[transition_pair] = [curr_time_sum]
                
                curr_time_sum = time_lengths[idx]

        return residence_times, transition_times

    def get_average_residence_times(self, hist_states, skip_init_time=None, skip_init_index=None):
        """Get the average residence time for each state in custom state encoding.

        Parameters
        ----------
        hist_states : list[object]
            The states calculated, should have the same dimensions as the `hist_*` arrays.
        skip_init_time : float, optional
            The amount of time skipped in the beginning, by default None
        skip_init_index : int, optional
            The number of indices skipped in the beginning, by default None

        Returns
        -------
        float
            The average residence time for each state, averaged over the times we transitioned into that state
        """
        residence_times, _ = self.analyze_custom_states(hist_states, skip_init_time, skip_init_index)

        residence_time_avgs = {}
        for s in residence_times:
            residence_time_avgs[s] = np.mean(residence_times[s])

        return residence_time_avgs

    def get_cold_start_prob(self):
        """Get the probability of cold start for the simulated trace.

        Returns
        -------
        float
            The probability of cold start calculated by dividing the number of cold start requests, over all requests
        """
        return self.total_cold_count / self.total_req_count


    def get_average_lifespan(self):
        """Get the average lifespan of each instance, calculated by the amount of time from creation of instance, until its expiration.

        Returns
        -------
        float
            The average lifespan
        """
        life_spans = np.array([s.get_life_span() for s in self.prev_servers])
        return life_spans.mean()

    
    def get_result_dict(self):
        """Get the results of the simulation as a dict, which can easily be integrated into web services.

        Returns
        -------
        dict
            A dictionary of different characteristics.
        """
        return {
            "reqs_cold": self.total_cold_count,
            "reqs_total": self.total_req_count,
            "reqs_warm": self.total_warm_count,
            "prob_cold": self.get_cold_start_prob(),
            "reqs_reject": self.total_reject_count,
            "prob_reject": self.total_reject_count / self.total_req_count,
            "lifespan_avg": self.get_average_lifespan(),
            "inst_count_avg": self.get_average_server_count(),
            "inst_running_count_avg": self.get_average_server_running_count(),
            "inst_idle_count_avg": self.get_average_server_idle_count(),
        }

    def print_trace_results(self):
        """Print a brief summary of the results of the trace.
        """
        self.calculate_time_lengths()

        print(f"Cold Starts / total requests: \t {self.total_cold_count} / {self.total_req_count}")
        print(f"Cold Start Probability: \t {self.total_cold_count / self.total_req_count:.4f}")

        print(f"Rejection / total requests: \t {self.total_reject_count} / {self.total_req_count}")
        print(f"Rejection Probability: \t\t {self.total_reject_count / self.total_req_count:.4f}")

        # average instance life span
        life_spans = np.array([s.get_life_span() for s in self.prev_servers])
        print(f"Average Instance Life Span: \t {life_spans.mean():.4f}")

        # average instance count
        print(f"Average Server Count:  \t\t {self.get_average_server_count():.4f}")
        # average running count
        print(f"Average Running Count:  \t {self.get_average_server_running_count():.4f}")
        # average idle count
        print(f"Average Idle Count:  \t\t {self.get_average_server_idle_count():.4f}")

    def trace_condition(self, t):
        """The condition for resulting the trace, we continue the simulation until this function returns false.

        Parameters
        ----------
        t : float
            current time in the simulation since the start of simulation

        Returns
        -------
        bool
            True if we should continue the simulation, false otherwise
        """
        return t < self.max_time

    @staticmethod
    def print_time_average(vals, probs, column_width=15):
        """Print the time average of states.

        Parameters
        ----------
        vals : list[object]
            The values for which the time average is to be printed
        probs : list[float]
            The probability of each of the members of the values array
        column_width : int, optional
            The width of the printed result for `vals`, by default 15
        """
        print(f"{'Value'.ljust(column_width)} Prob")
        print("".join(["="]*int(column_width*1.5)))
        for val, prob in zip(vals, probs):
            print(f"{str(val).ljust(column_width)} {prob:.4f}")

    def calculate_time_average(self, values, skip_init_time=None, skip_init_index=None):
        """calculate_time_average calculates the time-averaged of the values passed in with optional skipping a specific number of time steps (skip_init_index) and a specific amount of time (skip_init_time).

        Parameters
        ----------
        values : list
            A list of values with the same dimensions as history array (number of transitions)
        skip_init_time : Float, optional
            Amount of time skipped in the beginning to let the transient part of the solution pass, by default None
        skip_init_index : [type], optional
            Number of steps skipped in the beginning to let the transient behaviour of system pass, by default None

        Returns
        -------
        (list, list)
            returns (unq_vals, val_times) where unq_vals is the unique values inside the values list
            and val_times is the portion of the time that is spent in that value.
        """
        assert len(values) == len(self.time_lengths), "Values shoud be same length as history array (number of transitions)"

        skip_init = self.get_skip_init(skip_init_time=skip_init_time, 
                                        skip_init_index=skip_init_index)

        values = values[skip_init:]
        time_lengths = self.time_lengths[skip_init:]

        # get unique values
        unq_vals = list(set(values))
        val_times = []
        for val in unq_vals:
            t = time_lengths[[v == val for v in values]].sum()
            val_times.append(t)

        # convert to percent
        val_times = np.array(val_times)
        val_times = val_times / val_times.sum()
        return unq_vals, val_times

    def generate_trace(self, debug_print=False, progress=False):
        """Generate a sample trace.

        Parameters
        ----------
        debug_print : bool, optional
            If True, will print each transition occuring during the simulation, by default False
        progress : bool, optional
            Whether or not the progress should be outputted using the `tqdm` library, by default False

        Raises
        ------
        Exception
            Raises of FunctionInstance enters an unknown state (other than `IDLE` for idle or `TERM` for terminated) after making an internal transition
        """
        pbar = None
        if progress:
            pbar = tqdm(total=int(self.max_time))

        t = 0
        pbar_t_update = 0
        pbar_interval = int(self.max_time / 100)
        next_arrival = t + self.req()
        while self.trace_condition(t):
            if progress:
                if int(t - pbar_t_update) > pbar_interval:
                    pbar.update(int(t) - pbar_t_update)
                    pbar_t_update = int(t)
            self.hist_times.append(t)
            self.hist_server_count.append(self.server_count)
            self.hist_server_running_count.append(self.running_count)
            self.hist_server_idle_count.append(self.idle_count)
            if debug_print:
                print()
                print(f"Time: {t:.2f} \t NextArrival: {next_arrival:.2f}")
                print(self)
                # print state of all servers
                [print(s) for s in self.servers]

            # if there are no servers, next transition is arrival
            if self.has_server() == False:
                t = next_arrival
                next_arrival = t + self.req()
                # no servers, so cold start
                self.cold_start_arrival(t)
                continue

            # if there are servers, next transition is the soonest one
            server_next_transitions = np.array([s.get_next_transition_time(t) for s in self.servers])

            # if next transition is arrival
            if (next_arrival - t) < server_next_transitions.min():
                t = next_arrival
                next_arrival = t + self.req()

                # if warm start
                if self.idle_count > 0:
                    self.warm_start_arrival(t)
                # if cold start
                else:
                    self.cold_start_arrival(t)
                continue

            # if next transition is a state change in one of servers
            else:
                # find the server that needs transition
                idx = server_next_transitions.argmin()
                t = t + server_next_transitions[idx]
                new_state = self.servers[idx].make_transition()
                # delete instance if it was just terminated
                if new_state == 'TERM':
                    self.prev_servers.append(self.servers[idx])
                    self.idle_count -= 1
                    self.server_count -= 1
                    del self.servers[idx]
                    if debug_print:
                        print(f"Termination for: # {idx}")
                
                # if request has done processing (exit event)
                elif new_state == 'IDLE':
                    # transition from running to idle
                    self.running_count -= 1
                    self.idle_count += 1
                else:
                    raise Exception(f"Unknown transition in states: {new_state}")

        # after the trace loop, append the last time recorded
        self.hist_times.append(t)
        self.calculate_time_lengths()
        if progress:
            pbar.update(int(self.max_time) - pbar_t_update)
            pbar.close()
        




if __name__ == "__main__":
    sim = ServerlessSimulator(arrival_rate=0.9, warm_service_rate=1/2.016, cold_service_rate=1/2.163,
            expiration_threshold=600, max_time=100000)
    sim.generate_trace(debug_print=False, progress=True)
    sim.print_trace_results()
  


# ServerlessTemporalSimulator extends the functionality of ServerlessSimulator
# by providing necessary functionality for temporal scenarios where initial
# state is important and the initial process might be different from the
# following service process.

from pacssim.FunctionInstance import FunctionInstance
from pacssim.ServerlessSimulator import ServerlessSimulator


class ServerlessTemporalSimulator(ServerlessSimulator):
    """ServerlessTemporalSimulator extends ServerlessSimulator to enable extraction of temporal characteristics. Also gets all of the arguments accepted by :class:`~pacssim.ServerlessSimulator.ServerlessSimulator`

    Parameters
    ----------
    running_function_instances : list[FunctionInstance]
        A list containing the running function instances
    idle_function_instances : list[FunctionInstance]
        A list containing the idle function instances
    """
    def __init__(self, running_function_instances, idle_function_instances,
                 *args, **kwargs):

        super().__init__(*args, **kwargs)

        init_running_count = len(running_function_instances)
        init_idle_count = len(idle_function_instances)
        init_server_count = init_running_count + init_idle_count

        self.server_count = init_server_count
        self.running_count = init_running_count
        self.idle_count = init_server_count - init_running_count
        self.servers = [*running_function_instances, *idle_function_instances]



class ExponentialServerlessTemporalSimulator(ServerlessTemporalSimulator):
    """ExponentialServerlessTemporalSimulator is a simulator assuming exponential distribution for proceesing times which means each process is state-less and we can generate a service time and use that from now on. This class extends ServerlessTemporalSimulator which has functionality for other processes as well.

    Parameters
    ----------
    running_function_instance_count : integer
        running_function_instance_count is the number of instances currently processing a request
    idle_function_instance_next_terminations : list[float]
        idle_function_instance_next_terminations is an array of next termination scheduled for idle functions
        if they receive no new requests.
    """
    def __init__(self, running_function_instance_count, idle_function_instance_next_terminations,
                 *args, **kwargs):
        cold_service_process = ExpSimProcess(rate=cold_service_rate)
        warm_service_process = ExpSimProcess(rate=warm_service_rate)

        idle_functions = []
        for next_term in idle_function_instance_next_terminations:
            f = FunctionInstance(0,
                                cold_service_process,
                                warm_service_process,
                                expiration_threshold
                                )

            f.state = 'IDLE'
            f.is_cold = False
            # when will it be destroyed if no requests
            f.next_termination = next_term
            # so that they would be less likely to be chosen by scheduler
            f.creation_time = 0.01
            idle_functions.append(f)

        running_functions = []
        for _ in range(running_function_instance_count):
            f = FunctionInstance(0,
                                cold_service_process,
                                warm_service_process,
                                expiration_threshold
                                )

            f.state = 'IDLE'
            f.is_cold = False
            # transition it into running mode
            f.arrival_transition(0)

            running_functions.append(f)

        super().__init__(
            running_function_instances=running_functions,
            idle_function_instances=idle_functions,
            *args, **kwargs
        )


if __name__ == "__main__":
    from pacssim.SimProcess import ExpSimProcess

    print("Performing Temporal Simulation")
    cold_service_rate = 1/2.163
    warm_service_rate = 1/2.016
    expiration_threshold = 600

    arrival_rate = 0.9
    max_time = 600

    running_function_count = 3
    idle_function_count = 10

    cold_service_process = ExpSimProcess(rate=cold_service_rate)
    warm_service_process = ExpSimProcess(rate=warm_service_rate)

    idle_functions = []
    for _ in range(idle_function_count):
        f = FunctionInstance(0,
                             cold_service_process,
                             warm_service_process,
                             expiration_threshold
                             )

        f.state = 'IDLE'
        f.is_cold = False
        # when will it be destroyed if no requests
        f.next_termination = 300
        # so that they would be less likely to be chosen by scheduler
        f.creation_time = 0.01
        idle_functions.append(f)

    running_functions = []
    for _ in range(running_function_count):
        f = FunctionInstance(0,
                             cold_service_process,
                             warm_service_process,
                             expiration_threshold
                             )

        f.state = 'IDLE'
        f.is_cold = False
        # transition it into running mode
        f.arrival_transition(0)

        running_functions.append(f)

    sim = ServerlessTemporalSimulator(running_functions, idle_functions, arrival_rate=arrival_rate, warm_service_rate=warm_service_rate, cold_service_rate=cold_service_rate,
                                      expiration_threshold=expiration_threshold, max_time=max_time)
    sim.generate_trace(debug_print=False, progress=True)
    sim.print_trace_results()


    print("Testing out the new functionality added.")
    idle_function_instance_next_terminations = [300] * idle_function_count
    sim = ExponentialServerlessTemporalSimulator(running_function_count, idle_function_instance_next_terminations, arrival_rate=arrival_rate, 
                                        warm_service_rate=warm_service_rate, cold_service_rate=cold_service_rate,
                                        expiration_threshold=expiration_threshold, max_time=max_time)
    sim.generate_trace(debug_print=False, progress=True)
    sim.print_trace_results()



__version__ = '0.1.3-b'

__author__ = 'Nima Mahmoudi <nma@ieee.org>'

__all__ = []

# import package files
from . import SimProcess
from . import Utility
from . import ServerlessSimulator
from . import FunctionInstance





class FunctionInstance:
    """FunctionInstance aims to simulate the behaviour of a function instance in a serverless platform, with all the internal transitions necessary.

    Parameters
    ----------
    t : float
        The time at which the instance is being created
    cold_service_process : pacssim.SimProcess.SimProcess
        The process used to sample cold start response times
    warm_service_process : pacssim.SimProcess.SimProcess
        The process used to sample warm start response times
    expiration_threshold : float
        The amount of time it takes for an instance to get expired and the resources consumed by it released after processing the last request
    """
    def __init__(self, t, cold_service_process, warm_service_process, expiration_threshold):
        super().__init__()

        self.cold_service_process = cold_service_process
        self.warm_service_process = warm_service_process
        self.expiration_threshold = expiration_threshold

        # life span calculations
        self.creation_time = t

        # set current state variables
        self.state = 'COLD'
        self.is_busy = True
        self.is_cold = True

        # calculate departure and expected termination on each arrival
        self.next_departure = t + self.cold_service_process.generate_trace()
        self.update_next_termination()

    def __str__(self):
        return f"State: {self.state} \t Departure: {self.next_departure:8.2f} \t Termination: {self.next_termination:8.2f}"

    def get_life_span(self):
        """Get the lifespan of the instance (from creation until the termination)

        Returns
        -------
        float
            The number of seconds it took the instance to be expired
        """
        return self.next_termination - self.creation_time

    def update_next_termination(self):
        """Update the next scheduled termination if no other requests are made to the instance.
        """
        self.next_termination = self.next_departure + self.expiration_threshold

    def get_life_span(self):
        return self.next_termination - self.creation_time

    def get_state(self):
        return self.state

    def arrival_transition(self, t):
        """Make an arrival transition, which causes the instance to go from IDLE to WARM

        Parameters
        ----------
        t : float
            The time at which the transition has occured, this also updates the next termination.

        Raises
        ------
        Exception
            Raises if currently process a request by being in `COLD` or `WARM` states
        """
        if self.state == 'COLD' or self.state == 'WARM':
            raise Exception('instance is already busy!')

        elif self.state == 'IDLE':
            self.state = 'WARM'
            self.is_busy = True
            self.next_departure = t + self.warm_service_process.generate_trace()
            self.update_next_termination()

    def is_idle(self):
        """Whether or not the instance is currently idle, and thus can accept new requests.

        Returns
        -------
        bool
            True if idle, false otherwise
        """
        return self.state == 'IDLE'

    def make_transition(self):
        """Make the next internal transition, either transition into `IDLE` of already processing a request, or `TERM` if scheduled termination has arrived.

        Returns
        -------
        str
            The state after making the internal transition

        Raises
        ------
        Exception
            Raises if already in `TERM` state, since no other internal transitions are possible
        """
        # next transition is a departure
        if self.state == 'COLD' or self.state == 'WARM':
            self.state = 'IDLE'
            self.is_busy = False
            self.is_cold = False

        # next transition is a termination
        elif self.state == 'IDLE':
            self.state = 'TERM'
            self.is_busy = False

        # if terminated
        else:
            raise Exception("Cannot make transition on terminated instance!")

        return self.state

    def get_next_transition_time(self, t=0):
        """Get how long until the next transition.

        Parameters
        ----------
        t : float, optional
            The current time, by default 0

        Returns
        -------
        float
            The seconds remaining until the next transition
        """
        # next transition would be termination
        if self.state == 'IDLE':
            return self.get_next_termination(t)
        # next transition would be departure
        return self.get_next_departure(t)

    def get_next_departure(self, t):
        """Get the time until the next departure

        Parameters
        ----------
        t : float
            Current time

        Returns
        -------
        float
            Amount of time until the next departure

        Raises
        ------
        Exception
            Raises if called after the departure
        """
        if t > self.next_departure:
            raise Exception("current time is after departure!")
        return self.next_departure - t

    def get_next_termination(self, t):
        """Get the time until the next termination

        Parameters
        ----------
        t : float
            Current time

        Returns
        -------
        float
            Amount of time until the next termination

        Raises
        ------
        Exception
            Raises if called after the termination
        """
        if t > self.next_termination:
            raise Exception("current time is after termination!")
        return self.next_termination - t


import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import expon, poisson, norm

from pacssim.Utility import convert_hist_pdf

# import warnings
# warnings.simplefilter(action='ignore', category=FutureWarning)
# warnings.simplefilter(action='ignore', category=NotImplementedError)
# import modin.pandas as pd

class SimProcess:
    """SimProcess gives us a single interface to simulate different processes.
This will later on be used to simulated different processes and compare them agaist a
custom analytical model. In the child class, after performing `super().__init__()`,
properties `self.has_pdf` and `self.has_cdf` by default value of `False` will
be created. In case your class has the proposed PDF and CDF functions available,
you need to override these values in order for your model PDF to show up in the output
plot.
    """
    def __init__(self):
        super().__init__()
        # if your class has pdf or cdf functions, switch the booleans to True
        self.has_pdf = False
        self.has_cdf = False

    def pdf(self, x):
        """pdf function is called for visualization for classes with `self.has_pdf = True`.

        Parameters
        ----------
        x : float
            The time for which the pdf value (density) should be returned

        Raises
        ------
        NotImplementedError
            By default, this function raises NotImplementedError unless overriden by a child class.
        """
        raise NotImplementedError

    def cdf(self, x):
        """cdf function is called for visualization for classes with `self.has_cdf = True`.

        Parameters
        ----------
        x : float
            The time for which the cdf value (density) should be returned

        Raises
        ------
        NotImplementedError
            By default, this function raises NotImplementedError unless overriden by a child class.
        """
        raise NotImplementedError

    def generate_trace(self):
        """generate_trace function is supposed to be replaced with the override function of each
of the child classes.

        Raises
        ------
        NotImplementedError
            By default, this function raises NotImplementedError unless overriden by a child class.
        """
        raise NotImplementedError

    def visualize(self, num_traces=10000, num_bins=100):
        """visualize function visualizes the PDF and CDF of the simulated process by generating
traces from your function using :func:`~pacssim.SimProcess.SimProcess.generate_trace` and
converting the resulting histogram values (event counts) to densities to be comparable with
PDF and CDF functions calculated analytically.

        Parameters
        ----------
        num_traces : int, optional
            Number of traces we want to generate for calculating the histogram, by default 10000
        num_bins : int, optional
            Number of bins for the histogram which created the density probabilities, by default 100
        """
        traces = np.array([self.generate_trace() for i in range(num_traces)])
        print(f"Simulated Average Inter-Event Time: {np.mean(traces):.6f}")
        print(f"Simulated Average Event Rate: {num_traces / np.sum(traces):.6f}")

        base, hist_values, cumulative = convert_hist_pdf(traces, num_bins)

        plt.figure()
        plt.plot(base, hist_values, label='Sim Hist')
        if self.has_pdf:
            pdf_vals = np.array([0,0] + [self.pdf(x) for x in base[2:]])
            plt.plot(base, pdf_vals, ls='--', label="Model PDF")
        plt.legend()
        plt.grid(True)

        plt.figure()
        plt.plot(base, cumulative, label='Sim Cumulative')
        if self.has_cdf:
            cdf_vals = np.array([0, 0] + [self.cdf(x) for x in base[2:]])
            plt.plot(base, cdf_vals, ls='--', label="Model CDF")
        plt.legend()
        plt.grid(True)


class ExpSimProcess(SimProcess):
    """ExpSimProcess extends the functionality of :class:`~pacssim.SimProcess.SimProcess` for
exponentially distributed processes. This class also implements the `pdf` and `cdf` functions
which can be used for visualization purposes.

    Parameters
    ----------
    rate : float
        The rate at which the process should fire off
    """
    def __init__(self, rate):
        super().__init__()

        self.has_pdf = True
        self.has_cdf = True
        self.rate = rate

    def pdf(self, x):
        return expon.pdf(x, scale=1/self.rate)

    def cdf(self, x):
        return expon.cdf(x, scale=1/self.rate)

    def generate_trace(self):
        return np.random.exponential(1/self.rate)


class ConstSimProcess(SimProcess):
    """ConstSimProcess extends the functionality of :class:`~pacssim.SimProcess.SimProcess` for
constant processes, meaning this is a deterministic process and fires exactly every
`1/rate` seconds. This class does not implement the `pdf` and `cdf` functions.

    Parameters
    ----------
    rate : float
        The rate at which the process should fire off
    """
    def __init__(self, rate):
        super().__init__()

        self.has_pdf = False
        self.has_cdf = False
        self.rate = rate

    def generate_trace(self):
        return 1/self.rate


class GaussianSimProcess(SimProcess):
    """GaussianSimProcess extends the functionality of :class:`~pacssim.SimProcess.SimProcess` for
gaussian processes. This class also implements the `pdf` and `cdf` functions
which can be used for visualization purposes.

    Parameters
    ----------
    rate : float
        The rate at which the process should fire off
    std : float
        The standard deviation of the simulated process
    """
    def __init__(self, rate, std):
        super().__init__()
        self.has_pdf = True
        self.has_cdf = True
        self.rate = rate
        self.std = std

    def generate_trace(self):
        return max(0, np.random.normal(loc=1/self.rate, scale=self.std))

    def pdf(self, x):
        return norm.pdf(x, loc=1/self.rate, scale=self.std)

    def cdf(self, x):
        return norm.cdf(x, loc=1/self.rate, scale=self.std)

if __name__ == "__main__":
    print([ConstSimProcess(rate=5).generate_trace() for _ in range(10)])

    GaussianSimProcess(rate=5, std=0.01).visualize(num_traces=10000, num_bins=100)
    plt.show()

    exp_arr = ExpSimProcess(rate=5)
    exp_arr.visualize(num_traces=10000, num_bins=100)
    plt.show()



import setuptools
import re
import os
import ast

# parse version from __init__.py
_version_re = re.compile(r'__version__\s+=\s+(.*)')
_init_file = os.path.join(os.path.abspath(os.path.dirname(__file__)), "pacssim", "__init__.py")
with open(_init_file, 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

# Read the contents of requirements.txt
with open('requirements.txt') as content_file:
    content = content_file.read()

content = content.replace('\r', '')
reqs = content.split("\n")
reqs = [i for i in reqs if len(i) > 0]

setuptools.setup(
    name="pacssim",
    version=version,
    url="https://research.nima-dev.com",
    author="Nima Mahmoudi",
    author_email="nma@ieee.org",
    description="Developed in PACS Lab as a performance simulator for serverless computing platforms.",
    long_description=open('README.rst').read(),
    packages=setuptools.find_packages(),
    install_requires=reqs,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={},
)


import numpy as np

def convert_hist_pdf(_values, num_bins):
    """convert_hist_pdf converts the histogram resulting from _values and
num_bins to a density plot by dividing the probability of falling into a
bin by the bin size, converting the values to density. The resulting values
could be plotted and compared with the analytical pdf and cdf functions.

    Parameters
    ----------
    _values : list[float]
        A list of values that we want to analyze and calculate the histogram for
    num_bins : int
        Number of bins used for generating the histogram

    Returns
    -------
    list[float], list[float], list[float]
        base, values, cumulative are returned which are the histogram bases, density values, and cumulative densities which can be compared with the analytical cdf function
    """
    values, base = np.histogram(_values, bins=num_bins, density=False)
    bin_size = base[1] - base[0]
    cumulative = np.cumsum(values)
    cumulative = cumulative / cumulative[-1]
    base = np.append([0, base[0]], base[:-1])
    cumulative = np.append([0, 0], cumulative)
    values = np.append([0, 0], values)
    values = values / np.sum(values)
    if bin_size > 0:
        values /= bin_size

    base += bin_size/2

    return base, values, cumulative

